// ConvHull.inc
// version 1.0
//
// an include to create a convex hull of points using the 'beneath-beyond' algorithm
//
// by Michael Andrews, Feb 2003

#ifndef(_CONVEX_HULL_INC)
#declare _CONVEX_HULL_INC = version;
#version 3.5;

#ifdef(View_POV_Include_Stack)
   #debug "including convhull.inc\n"
#end

#include "math.inc"

// calculates dot product of a vector and normal of oriented triangle
#macro TrippleProd (_V, _A, _B, _C) vdot(vnormalize(_V), vnormalize(vcross(_B-_A,_A-_C))) #end

// Create convex hull of points using 'beneath-beyond' algorithm
//
// Pass in a point list P[], returns points, edges and triangles indices of convex hull in hP[], hE[] and hT[]
#macro Convex_Hull (P, hP, hE, hT)

  #local Nb = dimension_size(P, 1);

  // create hull list from an initial simplex
  #local Nh = 4;
  #local Hull = array[4][3] {{0,1,2},{0,2,3},{0,1,3},{1,2,3}}
  #local Ap = (P[0]+P[1]+P[2]+P[3])/4;

  #debug "Creating initial simplex ..."
  #local C = 0; #while (C < 4)
    #local Tc = (P[Hull[C][0]]+P[Hull[C][1]]+P[Hull[C][2]])/3;
    #local TProd = TrippleProd (Tc-Ap, P[Hull[C][0]], P[Hull[C][1]], P[Hull[C][2]]);
    #if (TProd < 0)
      #local TT = Hull[C][1];
      #declare Hull[C][1] = Hull[C][2];
      #declare Hull[C][2] = TT;
    #end
  #declare C = C + 1; #end
  #debug " done.\n"

  // for each new point
  #local C = 4; #while (C < Nb)
    #debug concat("Adding point ", str(C+1,0,0),"\r")
    #local Beyond = array[C]
    #local Horizon = array[C]

    #local C1 = 0; #while (C1 < C)
      #declare Beyond[C1] = false;
      #declare Horizon[C1] = false;
    #declare C1 = C1 + 1; #end

  //   create 'beyond' point list
  //   mark triangles for removal
    #local C1 = 0; #while (C1 < Nh)
      #local Tc = (P[Hull[C1][0]]+P[Hull[C1][1]]+P[Hull[C1][2]])/3;
      #local TProd = TrippleProd (P[C]-Tc, P[Hull[C1][0]], P[Hull[C1][1]], P[Hull[C1][2]]);
      #if (TProd >= 0)
        #declare Beyond[Hull[C1][0]] = true;
        #declare Beyond[Hull[C1][1]] = true;
        #declare Beyond[Hull[C1][2]] = true;
        #declare Hull[C1][0] = -1;
      #end
    #declare C1 = C1 + 1; #end

  //   remove triangles the point is 'beyond'
  //   set 'horizon' to all used hull points
    #local C1 = 0; #local C2 = 0; #while (C1 < Nh)
      #if (Hull[C1][0] > -1)
        #declare Hull[C2][0] = Hull[C1][0];
        #declare Hull[C2][1] = Hull[C1][1];
        #declare Hull[C2][2] = Hull[C1][2];
        #declare Horizon[Hull[C2][0]] = true;
        #declare Horizon[Hull[C2][1]] = true;
        #declare Horizon[Hull[C2][2]] = true;
        #declare C2 = C2 + 1;
      #end
    #declare C1 = C1 + 1; #end
    #declare Nh = C2;

  //   create 'horizon' point list - points in both 'horizon' and 'beyond'
  //   build average point from Hull points
    #local nHor = 0;

    #local Ap = x*0;
    #local nHull = 0;
    #local C1 = 0; #while (C1 < C)
      #if (Horizon[C1])
        #declare Ap = Ap + P[C1];
        #declare nHull = nHull + 1;
      #end
      #declare Horizon[C1] = (Horizon[C1] & Beyond[C1]);
      #if (Horizon[C1]) #declare nHor = nHor + 1; #end
    #declare C1 = C1 + 1; #end
    #declare Ap = Ap / nHull;

  //   if (horizon list is not empty)
    #if (nHor > 0)
      #local hList = array[nHor]
      #local C1 = 0; #local C2 = 0; #while (C1 < C)
        #if (Horizon[C1])
          #declare hList[C2] = C1;
          #declare C2 = C2 + 1;
        #end
      #declare C1 = C1 + 1; #end

  //     sort 'horizon' points into order around axis
      #local Axis = P[C]-Ap;
      #local V0 = vnormalize(VProject_Plane(P[hList[0]]-P[C], Axis));

      #local C1 = 1; #while (C1 < nHor - 1)
        #local C2 = C1; #while (C2 < nHor)
          #local V1 = vnormalize(VProject_Plane(P[hList[C1]]-P[C], Axis));
          #local V2 = vnormalize(VProject_Plane(P[hList[C2]]-P[C], Axis));
          #local A1 = VRotation(V1, V0, Axis); #if (A1 < 0) #declare A1 = 2*pi+A1; #end
          #local A2 = VRotation(V2, V0, Axis); #if (A2 < 0) #declare A2 = 2*pi+A2; #end
          #if (A2 < A1)
            #local tI = hList[C1];
            #declare hList[C1] = hList[C2];
            #declare hList[C2] = tI;
          #end
        #declare C2 = C2 + 1; #end
      #declare C1 = C1 + 1; #end

  //     increase Hull size by number of points in 'horizon'
      #local tHull = array[Nh+nHor][3]
      #local C1 = 0; #while (C1 < Nh)
        #declare tHull[C1][0] = Hull[C1][0];
        #declare tHull[C1][1] = Hull[C1][1];
        #declare tHull[C1][2] = Hull[C1][2];
      #declare C1 = C1 + 1; #end
      #declare Hull = tHull

  //     add new triangles to Hull
      #local C1 = 0; #while (C1 < nHor)
        #local C2 = mod(C1+1, nHor);
        #declare Hull[C1+Nh][0] = C;
        #declare Hull[C1+Nh][1] = hList[C1];
        #declare Hull[C1+Nh][2] = hList[C2];
      #declare C1 = C1 + 1; #end

      #declare Nh = Nh + nHor;

      #declare Ap = (Ap * C + P[C])/(C + 1);
    #end
  #declare C = C + 1; #end

  // set up Hull triangles, edges and points in return arrays
  #declare hT = Hull;
  #debug concat("\nHull: ", str(dimension_size(hT,1),0,0), " triangles set.\n")

  #local eTemp = array[Nb][Nb]
  #local pTemp = array[Nb]

  #local C1 = 0; #while (C1 < Nb-1)
    #local C2 = C1+1; #while (C2 < Nb)
      #declare eTemp[C1][C2] = false;
    #declare C2 = C2 + 1; #end
  #declare C1 = C1 + 1; #end

  #local C1 = 0; #while (C1 < Nb)
    #declare pTemp[C1] = false;
  #declare C1 = C1 + 1; #end

  #local C = 0; #local C1 = 0; #local C2 = 0; #while (C < Nh)
    #if (Hull[C][0] < Hull[C][1])
      #if (!eTemp[Hull[C][0]][Hull[C][1]]) #declare C1 = C1 + 1; #end
      #declare eTemp[Hull[C][0]][Hull[C][1]] = true;
    #end
    #if (Hull[C][1] < Hull[C][2])
      #if (!eTemp[Hull[C][1]][Hull[C][2]]) #declare C1 = C1 + 1; #end
      #declare eTemp[Hull[C][1]][Hull[C][2]] = true;
    #end
    #if (Hull[C][2] < Hull[C][0])
      #if (!eTemp[Hull[C][2]][Hull[C][0]]) #declare C1 = C1 + 1; #end
      #declare eTemp[Hull[C][2]][Hull[C][0]] = true;
    #end
    #if (!pTemp[Hull[C][0]]) #declare C2 = C2 + 1; #end
    #if (!pTemp[Hull[C][1]]) #declare C2 = C2 + 1; #end
    #if (!pTemp[Hull[C][2]]) #declare C2 = C2 + 1; #end
    #declare pTemp[Hull[C][0]] = true;
    #declare pTemp[Hull[C][1]] = true;
    #declare pTemp[Hull[C][2]] = true;
  #declare C = C + 1; #end

  #local Edges = array[C1][2]
  #local Points = array[C2]

  #local C = 0;
  #local C1 = 0; #while (C1 < Nb-1)
    #local C2 = C1+1; #while (C2 < Nb)
      #if (eTemp[C1][C2])
        #declare Edges[C][0] = C1;
        #declare Edges[C][1] = C2;
        #declare C = C + 1;
      #end
    #declare C2 = C2 + 1; #end
  #declare C1 = C1 + 1; #end

  #declare hE = Edges;
  #debug concat("Hull: ", str(dimension_size(hE,1),0,0), " edges set.\n")

  #local C = 0; #local C1 = 0; #while (C1 < Nb)
    #if (pTemp[C1])
      #declare Points[C] = C1;
      #declare C = C + 1;
    #end
  #declare C1 = C1 + 1; #end

  #declare hP = Points;
  #debug concat("Hull: ", str(dimension_size(hP,1),0,0), " points set.\n\n")

#end // end of Convex_Hull() macro

#version _CONVEX_HULL_INC;
#end // end _CONVEX_HULL_INC

